/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */

#include "safe_open.h"


/*******************************************************************************************
function:     safe_open
description:  Uses same input and output parameters as man(2) open.
              The filepath man(2) stat -ed, man (2) open -ed, and man (2) fstat -ed.
              The stat struct values from before the open are compared to the stat struct i 
              after the open.
              Any difference between the two sets of values will close the file descriptor.
              Somebody else has accessed the file in the meanwhile and you should assume 
              (un)deliberate tampering, hijacking and altering of the file.
result:
        < 0 : Failure state, errno is set appropriately, return value reflects where it went
              wrong in the code.
       >= 0 : Safely opened file descriptor.
*******************************************************************************************/
int safe_open (char * filepath, int flags, int mode)
{
    int fd = -1;
    int rc = 0; 
    struct stat stat_before, stat_after;

    char * errorbuffer = malloc (sizeof (char) * MAX_LOG_BUFFER_SIZE);
    if (!errorbuffer)
    {
        scas_log (1, "safe_open failure: Couldn't allocate error message buffer\n");
        return -121;
    } 

    /* 
     * Stat the filepath before open.
     * The Stat information from before the open must match the stat information after the open.
     * At the detection of a difference, close the file and label the attempt as tampered
     */

    if ((!filepath) || (strlen(filepath) == 0))
    {
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        scas_log (1, "safe_open initialization failure. %s", errorbuffer);
        rc = -10;
        goto safe_open_fail;
    }


    /*  Get the file status before opening the file */
    if (stat(filepath, &stat_before)) 
    {
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        scas_log (1, "Stat on file '%s' failed! %s\n", filepath, errorbuffer);
        rc = -20;
        goto safe_open_fail;
    }


    /* Open the file */
    if (!(fd = open(filepath, flags, mode)))
    {
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        scas_log (1, "Failed to open file '%s'. %s\n", filepath, errorbuffer);
        rc = -30;
        goto safe_open_fail;
    }

    /* Get file status of the opened file */
    if (fstat(fd, &stat_after))
    {
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        scas_log (1, "Stat on '%s' failed. %s\n", filepath, errorbuffer);
        rc = -40;
        goto safe_open_fail;
    }

    /* File status comparison between before and after the opening of the file */
    if ( (stat_before.st_dev     != stat_after.st_dev)     || 
         (stat_before.st_ino     != stat_after.st_ino)     ||
         (stat_before.st_mode    != stat_after.st_mode)    ||
         (stat_before.st_nlink   != stat_after.st_nlink)   ||
         (stat_before.st_rdev    != stat_after.st_rdev)    ||
         (stat_before.st_size    != stat_after.st_size)    ||
         (stat_before.st_blocks  != stat_after.st_blocks)  ||
         (stat_before.st_blksize != stat_after.st_blksize) ||
         /* Doesn't exist on Red Hat EL 3
          * (stat_before.st_flags   != stat_after.st_flags)   ||
          */
#ifndef _POSIX_C_SOURCE
      /* No need to compare the access times on the file (without modification) 
       * (stat_before.st_atimespec.tv_sec  != stat_after.st_atimespec.tv_sec)     ||
       * (stat_before.st_atimespec.tv_nsec != stat_after.st_atimespec.tv_nsec)    ||
       */
	/* Doesn't exist on OpenSolaris
	 * (stat_before.st_mtimespec.tv_sec  != stat_after.st_mtimespec.tv_sec)     ||
	 * (stat_before.st_mtimespec.tv_nsec != stat_after.st_mtimespec.tv_nsec)    ||
	 * (stat_before.st_ctimespec.tv_sec  != stat_after.st_ctimespec.tv_sec)     ||
	 * (stat_before.st_ctimespec.tv_nsec != stat_after.st_ctimespec.tv_nsec)    ||
	 */
#else
      /* No need to compare the access times on the file (without modification) 
       * (stat_before.st_atime      != stat_after.st_atime)    ||
       * (stat_before.st_atimesec   != stat_after.st_atimesec) ||
       */
         /* Doesn't exist on Red Hat EL 3
          * (stat_before.st_mtime      != stat_after.st_mtime)    ||
          * (stat_before.st_mtimesec   != stat_after.st_mtimesec) ||
          * (stat_before.st_ctime      != stat_after.st_ctime)    ||
          * (stat_before.st_ctimesec   != stat_after.st_ctimesec) ||
          */
#endif
	/* Might not work on RH EL3, but is POSIX following IEEE Std 1003.1-2001 */
	 (stat_before.st_mtime      != stat_after.st_mtime)    ||
	 (stat_before.st_ctime      != stat_after.st_ctime)    ||

         /* Doesn't exist on Red Hat EL 3
          * (stat_before.st_gen     != stat_after.st_gen)
          */

         (stat_before.st_uid     != stat_after.st_uid)     ||
         (stat_before.st_gid     != stat_after.st_gid) 
       )
    {
        close(fd);

        scas_log (1, "File status change detected during opening of the file. Possible hijack, tampering or dirty read of file '%s' prevented; closing file\n", filepath);
        rc = -111;
        goto safe_open_fail;
    }

    free(errorbuffer);
    return fd;

safe_open_fail:
    free(errorbuffer);
    return rc; 
}

